#version 410

#define pi 3.141593

uniform sampler2D _Diffuse_Map0; //diffuse albedo
uniform sampler2D _Normal_Map0;  //diffuse normal
uniform sampler2D _Diffuse_Map1; //specular albedo
uniform sampler2D _Normal_Map1;  //specular normal
uniform sampler2D _Disp_Map0;    //displacement map
uniform sampler2D _Shadow_Map0;  //shadow map
uniform sampler2D _Shadow_Map1;  //shadow map

uniform bool is_Diffuse_Map0;
uniform bool is_Normal_Map0;
uniform bool is_Diffuse_Map1;
uniform bool is_Normal_Map1;
uniform bool is_Disp_Map0;
uniform bool is_Shadow_Map0;
uniform bool is_Shadow_Map1;

uniform vec3 m_viewPos;
uniform vec3 m_lightPos0; //directional light
uniform vec3 m_lightColor0; //directional light
uniform vec3 m_lightPos1; //point light
uniform vec3 m_lightColor1; //point light

in vec2 uv_MainTex;
in vec3 worldPos;
in vec3 worldNormal;
in vec3 worldTangent;
in vec3 worldBiTangent;
in vec4 lightSpacePos0;
in vec4 lightSpacePos1;

out vec4 FragColor;

const float _Ka = 0.5; //ambient strength, (0, 1), 0.1
const float _Kd = 0.6; //diffuse strength, (0, 1), 0.1
const float _Ks = 0.4; //specular strength, (0, 1), 0.1

const vec3 SSSColor = vec3(0.75, 0.6, 0.5);
const bool is_Active = true;

vec4 gammaCorrection(vec4 vec, float g)
{
    return vec4(pow(vec.x, 1.0/g), pow(vec.y, 1.0/g), pow(vec.z, 1.0/g), vec.w);
}

vec3 getNormalMap(sampler2D sampler, vec2 uv)
{
    vec3 normalMap = texture(sampler, uv.st).xyz;
    // transform normal vector to range [-1,1]
    normalMap = normalize(normalMap * 2.0 - 1.0);
    normalMap = worldTangent * normalMap.x + worldBiTangent * normalMap.y + worldNormal * normalMap.z;
    normalMap = normalize(normalMap);
    return normalMap;
}

vec3 calcTangentNormal(sampler2D mesogeo)
{
    float mesoMapSize = 4096; //textureSize(mesogeo, 0).x // size of the displacement
    float ddMeso = 1.0/mesoMapSize;

    float dxx = texture(mesogeo, vec2(uv_MainTex.x+ddMeso, uv_MainTex.y)).r;
    float xxd = texture(mesogeo, vec2(uv_MainTex.x-ddMeso, uv_MainTex.y)).r;
    float dyy = texture(mesogeo, vec2(uv_MainTex.x, uv_MainTex.y+ddMeso)).r;
    float yyd = texture(mesogeo, vec2(uv_MainTex.x, uv_MainTex.y-ddMeso)).r;
    
    float dx = dxx - xxd;
    float dy = dyy - yyd;
    //vec3 n = normalize(cross(e1_world, e2_world));
    vec3 n = normalize(worldNormal); //normal
    vec3 t = normalize(worldTangent);
    vec3 bt = normalize(worldBiTangent);
    
    float scaleFactor = 1.0; // [FIXME] adjust this to control the bumpiness of the normal
    vec3 e1 = t*ddMeso*2.0 + n*dx*scaleFactor; //tangent
    vec3 e2 = bt*ddMeso*2.0 + n*dy*scaleFactor; //bi-tangent
    
    vec3 normal = normalize(cross(e1, e2));
    return normal;
}


float calcFresnel(float R0, float F, float fresExp, vec3 view, vec3 fspecNormal)
{
    return R0 + (F - R0) * pow(1.0 - dot(view, fspecNormal), fresExp);
}

float schlick(float r, float cos_theta)
{
    if(cos_theta<0.0){
        return 0.0;
    }
    return r+(1.0-r)*(pow(1.0-cos_theta,5.0));
}

float fresnelGraham(float r, float cos_theta, float power)
{
    if(cos_theta<0.0){
        return 0.0;
    }
    return r+(1.0-r)*(pow(1.0-cos_theta, power));
}

vec3 calcFresnel3(vec3 R0, float F, float fresExp, vec3 view, vec3 fspecNormal)
{
    return R0 + (vec3(F) - R0) * pow(1.0 - dot(view, fspecNormal), fresExp);
}

vec3 diffuseReflection(vec3 normal_R, vec3 normal_G, vec3 normal_B, vec3 L)
{
    vec3 dots = vec3(dot(normal_R, L), dot(normal_G, L), dot(normal_B, L));
    return pow(clamp((1.0f - dots) * SSSColor *0.5 + dots, 0.0, 1.0), SSSColor * 0.5 * 4.0f + 1.0f);
}


vec3 diffuseReflection2(vec3 normal_R, vec3 normal_G, vec3 normal_B, vec3 L)
{
    float ar = acos( dot( normal_R, L) );
    float ag = acos( dot( normal_G, L) );
    float ab = acos( dot( normal_B, L) );
    
    vec3 diff;
    
    diff.x = exp( -0.5*ar*ar - 0.1*ar*ar*ar*ar*ar*ar);
    diff.y = exp( -0.5*ag*ag - 0.1*ag*ag*ag*ag*ag*ag );
    diff.z = exp( -0.5*ab*ab - 0.1*ab*ab*ab*ab*ab*ab );
    
    return diff;
}

float geometric_attenuation(vec3 n, vec3 wi, vec3 wo)
{
    vec3 wh = (wi+wo);
    wh = normalize(wh);
    float NdotWh = abs(dot(n,wh));
    float NdotWo = abs(dot(n,wo));
    float NdotWi = abs(dot(n,wi));
    float WodotWh = abs(dot(wo,wh));
    if(WodotWh==0.0)
        return 0.0;
    
    return min(1.0, min((2.0 * NdotWh * NdotWo / WodotWh),
                        (2.0 * NdotWh * NdotWi / WodotWh)));
}

float microfacet_distribution(vec3 n, vec3 wh, float m)
{
    float NdotWh = (dot(n,wh));
    if(NdotWh<=0.0) return 0.0;
    if(NdotWh >= 1.0)
        NdotWh = 1.0;
    //float delta = acos(NdotWh);
    //float tan_delta = tan(delta);
    float tanByMSq = (NdotWh*NdotWh-1.0)/(NdotWh*NdotWh*m*m);
    //tanByMSq = tan(acos(NdotWh)) * tan(acos(NdotWh)) / (m*m);
    //float upper = exp(-1.0*(tan_delta/m)*(tan_delta/m));
    float upper = exp(tanByMSq);
    float lower = pi*m*m*NdotWh*NdotWh*NdotWh*NdotWh;
    return upper/lower;
    //return delta;
}


float ShadowCalculation(vec4 fragPosLightSpace, int id)
{
    // perform perspective divide
    vec3 projCoords = fragPosLightSpace.xyz / fragPosLightSpace.w;
    // transform to [0,1] range
    projCoords = projCoords * 0.5 + 0.5;
    // get closest depth value from light's perspective (using [0,1] range fragPosLight as coords)
    // float closestDepth = texture(_Shadow_Map0, projCoords.xy).r;
    // get depth of current fragment from light's perspective
    float currentDepth = projCoords.z;
    // calculate bias (based on depth map resolution and slope)
    vec3 normal = normalize(worldNormal);
    vec3 lightDir = id < 0 ? m_lightPos0 : m_lightPos1 ;
    float bias = max(0.0005 * (1.0 - dot(normal, lightDir)), 0.0001);
    // check whether current frag pos is in shadow
    // float shadow = currentDepth - bias > closestDepth  ? 1.0 : 0.0;
    // PCF
    float shadow = 0.0;
    vec2 texelSize = 1.0 / textureSize(id < 0 ?_Shadow_Map0 : _Shadow_Map1, 0);
    for(int x = -2; x <= 2; ++x)
    {
        for(int y = -2; y <= 2; ++y)
        {
            float pcfDepth = texture(id < 0 ?_Shadow_Map0 : _Shadow_Map1, projCoords.xy + vec2(x, y) * texelSize).r;
            shadow += currentDepth - bias > pcfDepth  ? 1.0 : 0.0;
        }
    }
    shadow /= 25.0;
    
    // keep the shadow at 0.0 when outside the far_plane region of the light's frustum.
    if(projCoords.z > 1.0)
        shadow = 0.0;
    
    return shadow;
}


void main(void)
{    
    float shadow = is_Shadow_Map0 ? ShadowCalculation(lightSpacePos0, -1) : 0;
    shadow += is_Shadow_Map1 ? ShadowCalculation(lightSpacePos1, 1) : 0;

    vec3 specNormal = normalize(worldNormal);
    // use displacement for specular normal
    if(is_Disp_Map0)
        specNormal = calcTangentNormal(_Disp_Map0);
    else if(is_Normal_Map1)
        specNormal = getNormalMap(_Normal_Map1, uv_MainTex);
    // diffuse normal map
    vec3 alpha;
    alpha.x = pow(1.0 + SSSColor.x, -5.0) - 1.0 / 32.0; // [NOTE] empirical fit
    alpha.y = pow(1.0 + SSSColor.y, -5.0) - 1.0 / 32.0; // [NOTE] empirical fit
    alpha.z = pow(1.0 + SSSColor.z, -5.0) - 1.0 / 32.0; // [NOTE] empirical fit
    vec3 diffNormalR = is_Normal_Map0 ? normalize(getNormalMap(_Normal_Map0, uv_MainTex)) : normalize(worldNormal);
    //diffNormalR = normalize(mix(diffNormalR, specNormal, alpha.r));
    vec3 diffNormalG = normalize(mix(diffNormalR, specNormal, alpha.g));
    vec3 diffNormalB = normalize(mix(diffNormalR, specNormal, alpha.b));
    
    vec4 alpha_channel = vec4(1.0, 1.0, 1.0, 1.0);
    vec4 diff_color = texture(_Diffuse_Map0, uv_MainTex);
    vec4 spec_color = is_Diffuse_Map1 ? texture(_Diffuse_Map1, uv_MainTex) : diff_color;

    vec3 pos = worldPos;
    vec3 view = m_viewPos;
    view = normalize(view - pos);
    vec3 light = m_lightPos0;
    light = normalize(light - pos);
    vec3 H = normalize(light + view);

    //ambient
    vec4 ambient = _Ka * alpha_channel * diff_color;

    //diffuse
    vec4 diff_reflection = vec4(0.0, 0.0, 0.0, 1.0);
    diff_reflection += vec4(diffuseReflection(diffNormalR, diffNormalG, diffNormalB, light), 1.0) * vec4(m_lightColor0, 1);
    vec4 diffuse = _Kd * alpha_channel * diff_color * diff_reflection;

    //specular
    //micro and meso
    float brdfMicro;
    float brdfMeso;
#if 0
    brdfMicro = specularReflection(_Normal_Map1, light, view, vec3(1, 1, 1));
    brdfMeso = specularReflection(_Normal_Map1, light, view, vec3(1, 1, 1));
#else
    float microBaseRoughness = 0.15;
    float mesoBaseRoughness = 0.25;
    //float F = schlick(0.04, dot(H, light));
    float F = fresnelGraham(0.05, dot(specNormal, view), 2.0);
    float G = geometric_attenuation(specNormal, light, view);
    //micro
    float D = microfacet_distribution(specNormal, H, microBaseRoughness);
    brdfMicro = D*G*F / (4.0*dot(specNormal, light)*dot(specNormal, view));
    //meso
    D = microfacet_distribution(specNormal, H, mesoBaseRoughness);
    brdfMeso = D*G*F / (4.0*dot(specNormal, light) * dot(specNormal, view));
#endif
    float brdf = 0.3 * brdfMicro + 0.7 * brdfMeso;
    vec4 specular_reflection = vec4(0,0,0,0);
    specular_reflection += brdf*dot(specNormal, light) * vec4(m_lightColor0, 1);
    vec4 specular = _Ks * alpha_channel * specular_reflection * spec_color;
    
    FragColor = ambient + (diffuse + specular) * (1-shadow);
    FragColor = vec4(clamp(FragColor.rgb, 0.0, 1.0), 1.0);
    
    //FragColor = vec4(specNormal.xyz, 1.0);
    //FragColor = vec4(diffNormalR.xyz, 1.0);// + 0.5 * vec4(1.0, 1.0, 1.0, 1.0);
    //FragColor = gammaCorrection(FragColor, 2.2);
}



